/*
******************************************************************************
@ScriptName: Extractor.ahk
@Author: Highend
@Created on: 07.09.2013 20:56:27
@Last modified: 2015-11-20 12:39:01
@Function: Extract, cleanup and delete archives
@Current version: v0.9.6
******************************************************************************
*/

#NoEnv
#NoTrayIcon
#SingleInstance off

; #SETTINGS# =================================================================
; Get the hwnd from the calling XYplorer's window
WinGet, callerHwnd, IDLast, A
callerHwnd := callerHwnd + 0

; Set .ini file
SplitPath, A_ScriptFullPath, , , ownFileExt
StringReplace, iniFile, A_ScriptFullPath, .%ownFileExt%, .ini

; Check for .ini file
IfNotExist, %iniFile%
	ShowProgress(".ini file not found, aborted!", 250, 3000)

; Get settings from .ini file
IniRead, ini_Application, %iniFile%, General, Application
IniRead, ini_CleanUpFolder, %iniFile%, General, CleanUpFolder
IniRead, ini_DeleteArchives, %iniFile%, General, DeleteArchives
IniRead, ini_DstInactivePane, %iniFile%, General, DstInactivePane

IniRead, ini_AppFilePath, %iniFile%, %ini_Application%, AppFilePath
IniRead, ini_ExtractArgs1, %iniFile%, %ini_Application%, ExtractArgs1
IniRead, ini_ExtractArgs2, %iniFile%, %ini_Application%, ExtractArgs2
IniRead, ini_Extensions, %iniFile%, %ini_Application%, Extensions

if (ini_Application = "ERROR")
	ShowProgress("Critical error: Check the content of the .ini file!", 400, 3000)

; Allow portable path usage (.\ | ..\)
if (RegExMatch(ini_AppFilePath, "^\."))
	ini_AppFilePath := RelToAbsPath(A_ScriptDir, ini_AppFilePath)

IfNotExist, %ini_AppFilePath%
{
	FileSelectFile, ini_AppFilePath, , ::{20d04fe0-3aea-1069-a2d8-08002b30309d}, Select the .exe file for %ini_Application%, *.exe
	if (ini_AppFilePath)
		IniWrite, %ini_AppFilePath%, %iniFile%, %ini_Application%, AppFilePath
	else
		ShowProgress("Extraction application not found, aborted!", 350, 2000)
}


; #MAIN PART# ================================================================
; Build list of command line arguments
Loop, %0%
	l_Args .= "`n" %A_Index%

; Remove first empty line
StringReplace, l_Args, l_Args, `n

; Put arguments into an array
StringSplit, a_Args, l_Args, `n
countArgs := a_Args0

; Put supplied list of selected files (in XYplorer) into an array
StringSplit, a_XYSelectedFiles, a_Args1, |
xyCountSelectedFiles := a_XYSelectedFiles0

; Loop over all supplied files
Loop, %xyCountSelectedFiles%
	fileListQuoted .= (A_Index = 1) ? a_XYSelectedFiles%A_Index% : """ """ a_XYSelectedFiles%A_Index%
; Quoting looks a bit weired...   

; Split first file
SplitPath, a_XYSelectedFiles1, xyFirstFileName, xyCurrentPath, xyFirstFileExt, xyFirstFileBaseName

dstPath := xyCurrentPath
; New base destination path if a second argument is given (inactive / not visible XY pane)
if (ini_DstInactivePane = 1 && countArgs = 2)
	dstPath := a_Args2

; Abort on wrong / unknown selections
if (countArgs < 1)
	ShowProgress("Please supply a '|' separated list of file(s), aborted!", 400, 2000)
if xyFirstFileExt not in %ini_Extensions%
	ShowProgress("You didn't supply a supported archive file, aborted!", 400, 2000)

; Check if xyFirstFileBaseName contains ".part<number>" || ".7z"
; TODO: Expand regex to include further archive formats
xyStrippedFirstFileBaseName := RegExReplace(xyFirstFileBaseName, "(.*?)(\.part\d+|\.7z).*", "$1")
xyExtractToFolderName := (xyStrippedFirstFileBaseName != xyFirstFileBaseName) ? xyStrippedFirstFileBaseName : xyFirstFileBaseName

; Single or multiple archives selected?
multipleArchives := (xyCountSelectedFiles > 1) ? 1 : 0

; Special requirements for the packer apps
if (multipleArchives) {

	; TODO: Add additional extraction tools
	MsgBox, 4132, , You have selected more than one file.`n`nClick "Yes" to extract everything into:`n"%dstPath%\%xyExtractToFolderName%"`n`nClick "No" to extract every archive into it's own folder.
	IfMsgBox Yes
	{
		; Multiple files, single folder
		singleExtractFolder := 1

		; These two apps can extract all items with a single call
		if (ini_Application = "ExtractNow")
			RunWait, "%ini_AppFilePath%" "%fileListQuoted%" %ini_ExtractArgs1% "%dstPath%\%xyExtractToFolderName%"
		else if (ini_Application = "Bandizip")
			RunWait, "%ini_AppFilePath%" %ini_ExtractArgs1% "%dstPath%\%xyExtractToFolderName%" "%fileListQuoted%"
		else
		{
			; All other applications need a loop to extract items one by one
			Loop, %xyCountSelectedFiles%
			{
				curFile := a_XYSelectedFiles%A_Index%

				if (ini_Application = "7-Zip")
					RunWait, "%ini_AppFilePath%" %ini_ExtractArgs1% "%curFile%" -o"%dstPath%\%xyExtractToFolderName%" %ini_ExtractArgs2%
				else if (ini_Application = "FreeArc")
					RunWait, "%ini_AppFilePath%" %ini_ExtractArgs1% "%curFile%" -dp"%dstPath%\%xyExtractToFolderName%"
				else if (ini_Application = "PACL")
					RunWait, "%ini_AppFilePath%" %ini_ExtractArgs1% -p"%dstPath%\%xyExtractToFolderName%" "%curFile%"
				else if (ini_Application = "WinRAR")
					RunWait, "%ini_AppFilePath%" %ini_ExtractArgs1% "%curFile%" "%dstPath%\%xyExtractToFolderName%\"
				else if (ini_Application = "WinZip")
					RunWait, "%ini_AppFilePath%" %ini_ExtractArgs1% "%curFile%" "%dstPath%\%xyExtractToFolderName%"
			}
		}
	}
	else
	{
		; Multiple files, multiple folders
		multiExtractFolder := 1

		; Loop over all supplied files
		Loop, %xyCountSelectedFiles%
		{
			curFile := a_XYSelectedFiles%A_Index%

			; Split current file
			SplitPath, curFile, , curFilePath, , curFileBaseName

			dstPath := curFilePath
			if (ini_DstInactivePane = 1 && countArgs = 2)
				dstPath := a_Args2

			; Store all destination folders base names (for later selection in XYplorer)
			multiFileBaseNames .= (A_Index = 1) ? curFileBaseName : "|" curFileBaseName

			if (ini_Application = "7-Zip")
				RunWait, "%ini_AppFilePath%" %ini_ExtractArgs1% "%curFile%" -o"%dstPath%\%curFileBaseName%" %ini_ExtractArgs2%
			else if (ini_Application = "Bandizip")
				RunWait, "%ini_AppFilePath%" %ini_ExtractArgs1% "%dstPath%\%curFileBaseName%" "%curFile%"
			else if (ini_Application = "ExtractNow")
				RunWait, "%ini_AppFilePath%" "%curFile%" %ini_ExtractArgs1% "%dstPath%\%curFileBaseName%"
			else if (ini_Application = "FreeArc")
				RunWait, "%ini_AppFilePath%" %ini_ExtractArgs1% "%curFile%" -dp"%dstPath%\%curFileBaseName%"
			else if (ini_Application = "PACL")
				RunWait, "%ini_AppFilePath%" %ini_ExtractArgs1% -p"%dstPath%\%curFileBaseName%" "%curFile%"
			else if (ini_Application = "WinRAR")
				RunWait, "%ini_AppFilePath%" %ini_ExtractArgs1% "%curFile%" "%dstPath%\%curFileBaseName%\"
			else if (ini_Application = "WinZip")
				RunWait, "%ini_AppFilePath%" %ini_ExtractArgs1% "%curFile%" "%dstPath%\%curFileBaseName%"
		}
	}
} else {
	; Single archive selected
	singleExtractFolder := 1

	if (ini_Application = "7-Zip")
		RunWait, "%ini_AppFilePath%" %ini_ExtractArgs1% "%a_XYSelectedFiles1%" -o"%dstPath%\%xyExtractToFolderName%" %ini_ExtractArgs2%
	else if (ini_Application = "Bandizip")
		RunWait, "%ini_AppFilePath%" %ini_ExtractArgs1% "%dstPath%\%xyExtractToFolderName%" "%a_XYSelectedFiles1%"
	else if (ini_Application = "ExtractNow")
		RunWait, "%ini_AppFilePath%" "%a_XYSelectedFiles1%" %ini_ExtractArgs1% "%dstPath%\%xyExtractToFolderName%"
	else if (ini_Application = "FreeArc")
		RunWait, "%ini_AppFilePath%" %ini_ExtractArgs1% "%a_XYSelectedFiles1%" -dp"%dstPath%\%xyExtractToFolderName%"
	else if (ini_Application = "PACL")
		RunWait, "%ini_AppFilePath%" %ini_ExtractArgs1% -p"%dstPath%\%xyExtractToFolderName%" "%a_XYSelectedFiles1%"
	else if (ini_Application = "WinRAR")
		RunWait, "%ini_AppFilePath%" %ini_ExtractArgs1% "%a_XYSelectedFiles1%" "%dstPath%\%xyExtractToFolderName%\"
	else if (ini_Application = "WinZip")
		RunWait, "%ini_AppFilePath%" %ini_ExtractArgs1% "%a_XYSelectedFiles1%" "%dstPath%\%xyExtractToFolderName%"
}

; Cleanup after extraction
if (ini_CleanUpFolder = 1) {
	; In a single or multiple destination folders
	Loop, %xyCountSelectedFiles%
	{
		if (singleExtractFolder = 1) {
			dstPath := xyCurrentPath
			curFileBaseName := xyExtractToFolderName
			if (ini_DstInactivePane = 1 && countArgs = 2)
				dstPath := a_Args2
		} else if (multiExtractFolder = 1) {
			curFile := a_XYSelectedFiles%A_Index%
			SplitPath, curFile, , dstPath, , curFileBaseName
			if (ini_DstInactivePane = 1 && countArgs = 2)
				dstPath := a_Args2
		}

		; Count items (files & folders)
		countFile := 0
		Loop, %dstPath%\%curFileBaseName%\*.*, 1
		{
			if A_LoopFileAttrib not contains D
			{
				countFile := 1
				break
			}
			else
			{
				countFolders := A_Index
				moveFolderName := A_LoopFileName
			}
		}
		; Only move a single existing subfolder one level up if no other file exist!
		if (countFolders == 1 && countFile == 0)
			FileMoveDir, %dstPath%\%curFileBaseName%\%moveFolderName%, %dstPath%\%curFileBaseName%, 2
	}
}

; Delete archive file(s)
if (ini_DeleteArchives = 1) {
	if (multipleArchives = 1) {
		; Loop over all supplied files
		Loop, %xyCountSelectedFiles%
		{
			fileName := a_XYSelectedFiles%A_Index%
			l_filesToSelect .= (A_Index = 1) ? fileName : "`n" fileName
		}
	} else {
		; The following patterns try to match only multipart archives
		; New rar archive format: <name>.part<number>.rar (all)
		; Old rar archive format: <name>.rar | <name>.r<number> (following)
		; New zip archive format: <name>.zipx | <name>.zx<number> (following)
		; Old zip archive format: <name>.zip | <name>.z<number> (following)
		; 7-Zip archive format: <name>.7z.<number> (all)
		rarNewFilterPattern   := RegExReplace(xyFirstFileName, "^(.*?\.part)\d+\.rar$", "$1")
		rarOldFilterPattern   := RegExReplace(xyFirstFileName, "^((?!.+\.part\d+\.).+\.r)ar", "$1")
		zipNewFilterPattern   := RegExReplace(xyFirstFileName, "^(.*?\.z)(ip)(x)$", "$1$3")
		zipOldFilterPattern   := RegExReplace(xyFirstFileName, "^(.*?\.z)ip$", "$1")
		sevenZipFilterPattern := RegExReplace(xyFirstFileName, "^(.*?\.7z\.)\d+$", "$1")

		;MsgBox, Patterns found:`n`nrar (new): %rarNewFilterPattern%`nrar   (old): %rarOldFilterPattern%`nzip (new): %zipNewFilterPattern%`nzip   (old): %zipOldFilterPattern%`n7-zip       : %sevenZipFilterPattern%

		if (rarNewFilterPattern != xyFirstFileName)
			filterPattern := rarNewFilterPattern
		else if (rarOldFilterPattern != xyFirstFileName)
			filterPattern := rarOldFilterPattern
		else if (zipNewFilterPattern != xyFirstFileName)
			filterPattern := zipNewFilterPattern
		else if (zipOldFilterPattern != xyFirstFileName)
			filterPattern := zipOldFilterPattern
		else if (sevenZipFilterPattern != xyFirstFileName)
			filterPattern := sevenZipFilterPattern
		else
			; Different archiv type (and no multipart archive) -> use first file only
			filterPattern := xyFirstFileName

		; Build initial list
		Loop, %xyCurrentPath%\%filterPattern%*, 0
			l_filesToSelect .= (A_Index = 1) ? A_LoopFileLongPath : "`n" A_LoopFileLongPath

		; Exceptions for old zip multipart archives
		if (xyFirstFileExt = "zip") {
			Loop, parse, l_filesToSelect, `n
			{
				SplitPath, A_LoopField, , , selExt
				if selExt contains zipx,zx
					continue
				l_tempFilesToSelect .= (A_Index = 1) ? A_LoopField : "`n" A_LoopField
			}
			l_filesToSelect := l_tempFilesToSelect
		}

		if (xyFirstFileExt = "zipx") {
			if (l_filesToSelect)
				; Multipart archive
				l_filesToSelect := xyCurrentPath "\" xyFirstFileName "`n" l_filesToSelect
			else
				; Singlepart archive
				l_filesToSelect := xyCurrentPath "\" xyFirstFileName
		}

	}
}
else
	ExitApp

MainGUI:
; +Owner = no taskbar button
Gui, +AlwaysOnTop +Resize +Owner
Gui, Add, ListView, x0 y0 Grid Checked vFileListView, File
Gui, Add, Checkbox, Checked vToggleSelection gToggleSelection, &Toggle
Gui, Add, Button, w100 h30 Default vButtonDelete gButtonDelete, &Delete file(s)
Gui, Add, Button, w100 h30 vButtonCancel gButtonCancel, &Cancel

Menu, ContextMenu, Add, Check all, ContextCheckAll
Menu, ContextMenu, Add, Uncheck all, ContextUncheckAll
Menu, ContextMenu, Default, Uncheck all

PopulateListView(l_filesToSelect)
Gui, Show, w600 h400
return

ToggleSelection:
rowNumber := LV_GetNext(0, "Checked")
if not rowNumber
	LV_Modify(0, "Check")
else
	LV_Modify(0, "-Check")
return

ButtonDelete:
rowNumber := 0
l_checkedFiles := ""
Loop
{
	rowNumber := LV_GetNext(rowNumber, "Checked")
	if not rowNumber
		break
	LV_GetText(fileName, rowNumber)
	l_checkedFiles .= (A_Index = 1) ? fileName : "|" fileName
}

xyScript =
( LTrim Join
	::$filesToDelete = "%l_checkedFiles%";
	if ($filesToDelete) {
		delete 1, 0, "$filesToDelete";
	}
	$curPath = get("path");
	if (("$curPath" == "%xyCurrentPath%") && ($filesToDelete)) {
		if (%multiExtractFolder%){
			selectitems "%multiFileBaseNames%";
		} else {
			selectitems "%curFileBaseName%";
		}
	}
)

Gui, Hide
; Send to XYplorer only if it's corresponding window is the active one
IfWinActive, ahk_id %callerHwnd%
	Send_WM_COPYDATA(xyScript)
else
{
	; Otherwise just delete the files
	Loop, parse, l_checkedFiles, |
		FileRecycle, %A_LoopField%
}
ExitApp
return

ButtonCancel:
ExitApp
return

GuiContextMenu: ; Launched in response to a right-click or press of the Apps key.
if A_GuiControl <> FileListView ; Display the menu only for clicks inside the ListView.
	return
; Show the menu at the provided coordinates, A_GuiX and A_GuiY. These should be used
; because they provide correct coordinates even if the user pressed the Apps key.
Menu, ContextMenu, Show, %A_GuiX%, %A_GuiY%
return

ContextCheckAll:
LV_Modify(0, "Check")
return

ContextUncheckAll:
LV_Modify(0, "-Check")
return

GuiSize:
if A_EventInfo = 1 ; The window has been minimized. No action needed.
	return
; Otherwise, the window has been resized or maximized. Resize the ListView to match.
GuiControl, Move, FileListView, % "W" (A_GuiWidth - 1) "H" (A_GuiHeight - 40)
GuiControl, MoveDraw, ToggleSelection, % "X" (A_GuiWidth - 65) "Y" (A_GuiHeight - 25)
GuiControl, MoveDraw, ButtonDelete, % "X" ((A_GuiWidth - 205) / 2) "Y" (A_GuiHeight - 35)
GuiControlGet, ButtonDelete, Pos
GuiControl, MoveDraw, ButtonCancel, % "X" (ButtonDeleteX + 105) "Y" (A_GuiHeight - 35)
LV_ModifyCol(1, "AutoHdr")
return

GuiClose:
ExitApp
return

GuiEscape:
ExitApp
return

; #FUNCTIONS# ================================================================
; Show progress status
ShowProgress(message, width=300, time=1500, quit=1)
{
	Progress, B1 zh0 cwFFFFFF ctFF0000 fs12 w%width%, %message%
	Sleep, %time%
	if quit
		ExitApp
}

; Populate the ListView
PopulateListView(listData, separator="`n")
{
	GuiControl, -Redraw, ListView
	LV_Delete()

	Loop, parse, listData, %separator%
	{
		LV_Add("Check", A_LoopField)
	}

	GuiControl, +Redraw, ListView
	return
}

Send_WM_COPYDATA(message)
{
	global callerHwnd

	size := StrLen(message)
	if !(A_IsUnicode) {
		VarSetCapacity(data, size * 2, 0)
		StrPut(message, &data, size, "UTF-16")
	} else {
		data := message
	}
	VarSetCapacity(COPYDATA, A_PtrSize * 3, 0)
	NumPut(4194305, COPYDATA, 0, "Ptr")
	NumPut(size * 2, COPYDATA, A_PtrSize, "UInt")
	NumPut(&data, COPYDATA, A_PtrSize * 2, "Ptr")
	;SendMessage, 0x4a, 0, &COPYDATA, , ahk_id %callerHwnd%
	result := DllCall("User32.dll\SendMessageW", "Ptr", callerHwnd, "UInt", 74, "Ptr", 0, "Ptr", &COPYDATA, "Ptr")
	return
}

; http://www.autohotkey.com/board/topic/17922-func-relativepath-absolutepath/page-2#entry117257
; Convert relative path to absolute path
RelToAbsPath(d1, d2, s = "\") {
	d0 := SubStr(d1, 1, len := InStr(d1, s, "", InStr(d1, s . s) + 2) - 1)
		, d1 := SubStr(d1, len + 1), off := 0
	Loop, 2
		If InStr(d%A_Index%, s, "", 0) = StrLen(d%A_Index%)
			StringTrimRight, d%A_Index%, d%A_Index%, 1
	Loop, Parse, d2, %s%
		If InStr(A_LoopField, "..") = 1
			d1 := SubStr(d1, 1, InStr(d1, s, "", 0) - 1)
		Else If InStr(A_LoopField, ".") = 1
			StringTrimLeft, d2, d2, 2
		Else If A_LoopField =
		{
			d1 =
			StringTrimLeft, d2, d2, 1
		}
	StringReplace, d2, d2, ..%s%, , All
	Return, d0 . d1 . s . d2
}

/*
******************************************************************************
@Changelog:
v0.9.7:
  * Fixed a bug with CleanUpFolder = true when the destination folder contains not only one subfolder
    but at least one additional file
  * New icon by courtesy of kunkel321

v0.9.6:
  * Fixed the GUI bug that files aren't visible when DstInactivePane=1
  * Fixed a bug for destination folder cleanup when CleanUpFolder=1
  * Minor changes to the xy script

v0.9.5:
  * Added a new .ini option: DstInactivePane
    If set to 1: Extract all archives to the inactive (or if dual pane is not active: hidden) pane!
    If set to 0 (default): Extract to the base folder of the archive (old behavior)
    New call: run """<path to Extractor.exe>"" ""<get SelectedItemsPathNames |>"" ""<get Path i>""";

v0.9.4:
  * Added support for relative paths (.\ | ..\)
  * Added PACL (command line utility from PowerArchiver) support

v0.9.3:
  * Fixed old rar multipart detection (support for all file names, new rar multipart archives are excluded)
  * Fallback for different archive types

v0.9.2:
  * Fixed detection of single part zipx archives
  * Added a line for not supported archives for each application in Extractor.ini
  * Checked supported archives types for each application

v0.9.1:
  + Cleanup even multiple extraction folders
  + Delete archives inside XYplorer to retain undo functionality
  + Added more extraction tools (FreeArc, WinZip)
  + Resizable GUI (for deleting archives)
******************************************************************************
*/
